#!/usr/bin/env python
# coding=utf8

from logbook import Logger

from exc import ParserError

class BaseParser(object):
    @property
    def lineno(self):
        return self.tokenizer.lineno

    @property
    def colno(self):
        return self.tokenizer.colno

    def _expect(self, *args):
        t = self.tokens.next()
        if not args == t[:len(args)]:
            raise ParserError(self, 'Expected %s, got %s instead' % \
                (args, t)
            )
        return t


class BlockParser(BaseParser):
    BLOCK_TYPES = ['LC_IDENTIFICATION', 'LC_CTYPE', 'LC_COLLATE', 'LC_TIME',
                   'LC_NUMERIC', 'LC_MONETARY', 'LC_MESSAGES', 'LC_PAPER',
                   'LC_NAME', 'LC_ADDRESS', 'LC_TELEPHONE']
    log = Logger('BlockParser')

    def __init__(self, tokenizer):
        self.tokenizer = tokenizer
        self.tokens = iter(self.tokenizer)

    def parse(self):
        def _yield_until_end_of(blocktype):
            prev = None
            while True:
                t = self.tokens.next()
                if ('KEYWORD', 'END') == t and ('EOL',) == prev:
                    self._expect('KEYWORD', blocktype)
                    self._expect('EOL')
                    # done with the section
                    raise StopIteration

                yield t
                prev = t

        for t in self.tokens:
            if t[0] == 'KEYWORD' and t[1] in self.BLOCK_TYPES:
                handler = getattr(self, 'handle_' + t[1], None)
                gen = _yield_until_end_of(t[1])

                if not handler:
                    for g in gen:
                        pass  # skip until end
                    self.log.debug('skipping %s' % t[1])
                else:
                    handler(gen)


class TranslitParser(BlockParser):
    log = Logger('TranslitParser')
    def __init__(self, parse_func, *args, **kwargs):
        super(TranslitParser, self).__init__(*args, **kwargs)
        self.parse_func = parse_func
        self.ttbl = {}

        self.has_LC_IDENITIFCATION = False

    def handle_LC_IDENTIFICATION(self, tokens):
        self.has_LC_IDENITIFCATION = True
        for t in tokens:
            pass

    def handle_LC_CTYPE(self, tokens):
        self.log.debug('enter LC_CTYPE')
        inside_translit = False
        token_iter = iter(tokens)
        for t in token_iter:
            if ('KEYWORD', 'copy') == t:
                if inside_translit:
                    raise ParserError(self, 'copy inside translit section')

                t_fn = token_iter.next()
                if not (t_fn[0], 'STRING'):
                    raise ParserError(self, 'expected filename to copy')

                fn = t_fn[1]
                self.log.info('copying file "%s"' % fn)
                self.ttbl.update(self.parse_func(fn))

                if not ('EOL',) == token_iter.next():
                    raise ParserError(self, 'garbage after copy')
                continue

            if ('KEYWORD', 'include') == t:
                if not inside_translit:
                    raise ParserError(self, 'include outside translit section')

                t_fn = token_iter.next()
                if not (t_fn[0], 'STRING'):
                    raise ParserError(self, 'expected filename to copy')

                fn = t_fn[1]
                self.log.info('including file "%s"' % fn)
                self.ttbl.update(self.parse_func(fn))

                while token_iter.next() != ('EOL',):
                    pass
                continue

            if ('KEYWORD', 'translit_start') == t:
                self.log.debug('enter translit')
                inside_translit = True
                continue

            if ('KEYWORD', 'translit_end') == t:
                self.log.debug('exit translit')
                inside_translit = False
                continue

            if ('KEYWORD', 'default_missing') == t:
                self.log.debug('skipping default_missing')
                while token_iter.next() != ('EOL',):
                    pass
                continue

            if inside_translit and t != ('EOL',):
                # "ordinary" translit line
                groups = []

                current_group = []
                while t != ('EOL',):
                    if ('SEMICOLON',) == t:
                        groups.append(current_group)
                        current_group = []
                        t = token_iter.next()
                        continue

                    if not 'STRING' == t[0]:
                        raise ParserError(self, 'string expected, found %r'\
                                                'instead' % (t,))
                    current_group.append(t[1])

                    t = token_iter.next()

                if current_group:
                    groups.append(current_group)

                if len(groups) == 1:
                    assert len(groups[0]) == 2
                    lhs = [groups[0][0]]
                    rhs = groups[0][1]
                elif len(groups) >= 2:
                    # last substition seems to be the simplest
                    lhs = groups[0]
                    rhs = groups[-1][0]

                    assert len(groups[-1]) == 1

                    if len(groups) > 2:
                        self.log.warning('Unhandled leftover substitutions: '\
                                         '%s' % groups)

                for l in lhs:
                    self.ttbl[l] = rhs


        if inside_translit:
            raise ParserError(self, 'translit_end not found')
        self.log.debug('exit LC_CTYPE')



#!/usr/bin/env python
# coding=utf8


class LException(Exception):
    def __init__(self, src, err):
        super(LException, self).__init__(err)

        self.src = src
        self.err = err


class TokenizerError(LException):
    pass


class ParserError(LException):
    pass



#!/usr/bin/env python
# coding=utf8

import re

PREFACE_RE = re.compile(
    '^(\s*(?:escape_char|comment_char)\s*.\s*\n)*'
)

def strip_comments(comment_char, escape_char, string_delim, i):
    inside_string = False
    escaped = False

    # remove comments
    while True:
        c = i.next()

        if escaped:
            escaped = False
            yield c
            continue

        if escape_char == c:
            escaped = True
        elif string_delim == c:
            inside_string = not inside_string
        elif comment_char == c and not inside_string:
            while c != '\n':
                c = i.next()

        yield c


def preprocess(buf):
    preface_m = PREFACE_RE.search(buf)

    escape_char = '\\'
    comment_char = '#'
    string_delim = '"'

    if preface_m:
        for line in preface_m.group(0).split('\n'):
            # skip blank lines
            if not line:
                continue

            cmd, ch = line.strip().split()

            if cmd == 'escape_char':
                escape_char = ch
            elif cmd == 'comment_char':
                comment_char = ch

        # start reading after preface
        i = iter(buffer(buf, preface_m.end()))
    else:
        i = iter(buf)

    return escape_char,\
           string_delim,\
           strip_comments(comment_char,
                          escape_char,
                          string_delim,
                          i)


class Screener(object):
    def __init__(self, buf):
        self.escape_char, self.string_delim, self.i = preprocess(buf)
        self.lineno = 0
        self.colno = 0
        self.inside_string = False
        self._newline = False
        self.escaped = False

    def __iter__(self):
        def _getch():
            if self._newline:
                self.lineno += 1
                self.colno = 0
                self._newline = False
            else:
                self.colno += 1

            c = self.i.next()
            if '\n' == c:
                self._newline = True
            return c

        # skip initial whitespace
        c = _getch()
        while c.isspace():
            c = _getch()

        while True:
            if self.escaped:
                if '\n' != c:
                    # escaped newlines simply disappear
                    yield self.escape_char
                    yield c
                c = _getch()
                self.escaped = False
                continue

            # not escaped:
            if self.escape_char == c:
                self.escaped = True
                c = _getch()
                continue
            elif self.string_delim == c:
                self.inside_string = not self.inside_string
                c = _getch()
                yield self.string_delim
                continue
            if not self.inside_string:
                if '\n' == c:
                    # compact newlines
                    while '\n' == c:
                        c = _getch()
                    yield '\n'
                    continue
                elif c.isspace():
                    while c.isspace() and c != '\n':
                        c = _getch()
                    yield ' '
                    continue

            yield c
            c = _getch()
            continue



#!/usr/bin/env python
# coding=utf8

import re

try:
    from cStringIO import StringIO
except ImportError:
    from StringIO import StringIO

from exc import TokenizerError

KEYWORDS = ('LC_IDENTIFICATION', 'LC_CTYPE', 'LC_COLLATE', 'LC_TIME',
            'LC_NUMERIC', 'LC_MONETARY', 'LC_MESSAGES', 'LC_PAPER', 'LC_NAME',
            'LC_ADDRESS', 'LC_TELEPHONE', 'END', 'escape_char', 'comment_char',
            'translit_start', 'translit_end', 'include', 'copy', 'IGNORE',
            'default_missing')

_U_RE = re.compile('<U([0-9A-F]+)>')
def _uni_sub(s):
    def _u_repl(m):
        code_point = int(m.group(1), 16)
        return unichr(code_point)

    return _U_RE.sub(_u_repl, s)

class Tokenizer(object):
    def __init__(self, scrn):
        self.scrn = scrn
        self.word = StringIO()

    @property
    def lineno(self):
        return self.scrn.lineno

    @property
    def colno(self):
        return self.scrn.colno

    @property
    def escape_char(self):
        return self.scrn.escape_char

    @property
    def inside_string(self):
        return self.scrn.inside_string

    @property
    def escaped(self):
        return self.scrn.escaped

    def __iter__(self):
        def _getch():
            c = i.next()
            if self.escape_char == c:
                # drop escape chars
                return _getch()
            return c

        def _reset_word():
            val = self.word.getvalue()
            self.word = StringIO()
            return val

        def _append_char(c):
            self.word.write(c)

        i = iter(self.scrn)
        _reset_word()

        c = _getch()
        while True:
            # non-escaped string start
            if '"' == c and self.inside_string:
                c = _getch()  # skip initial "
                while self.inside_string:
                    if '\n' == c:
                        raise TokenizerError(self, "EOL inside string")

                    _append_char(c)
                    c = _getch()
                c = _getch()  # skip final "
                yield ('STRING', _uni_sub(_reset_word()))
                continue

            # yield EOLs
            if '\n' == c:
                yield ('EOL',)
                c = _getch()
                continue

            # skip whitespace
            if ' ' == c:
                c = _getch()
                continue

            if c == ';':
                yield ('SEMICOLON',)
                c = _getch()
                continue

            if c.isdigit():
                while c.isdigit():
                    _append_char(c)
                    c = _getch()

                yield ('INTEGER', int(_reset_word()))
                continue

            # reads a word
            while self.escaped or c not in (';', ' ', '\n', '"'):
                _append_char(c)
                c = _getch()

            w = _reset_word()
            if w in KEYWORDS:
                yield ('KEYWORD', w)
            else:
                yield ('STRING', _uni_sub(w))



#!/usr/bin/env python
# coding=utf8

import os
import sys

from setuptools import setup, find_packages

def read(fname):
    return open(os.path.join(os.path.dirname(__file__), fname)).read()


setup(name='slugger',
      version='0.1',
      description="""Slugging done right. Tries to support close to 300
                     languages.""",
      long_description=read('README.rst'),
      keywords='slug slugging web i18n',
      author='Marc Brinkmann',
      author_email='git@marcbrinkmann.de',
      url='http://github.com/mbr/slugger',
      license='LGPL',
      install_requires=['remember', 'logbook', 'unihandecode'],
      include_package_data=True,
      packages=find_packages(exclude=['glibcparse']),
      classifiers=[
          'Development Status :: 4 - Beta',
          'Environment :: Web Environment',
          'Intended Audience :: Developers',
          'License :: OSI Approved :: GNU Library or Lesser General Public License (LGPL)',
          'Operating System :: OS Independent',
          'Topic :: Internet',
          'Topic :: Internet :: WWW/HTTP :: Dynamic Content :: CGI Tools/Libraries',
          'Topic :: Software Development :: Internationalization',
     ]
     )



#!/usr/bin/env python
# coding=utf8

"""A script to extract glibc localedata.

Woefully under documented. All its components are found in the glibcparse
package. If you need a parser for glibc locale files for some reason, this is
not a bad starting point. Message me on <https://github.com/mbr/slugger> for
help.
"""

import bz2
import os
import sys

import logbook
import logbook.more

try:
    import cPickle as pickle
except ImportError:
    import pickle

try:
    from remember.memoize import memoize
    has_memoize = True
except ImportError:
    has_memoize = False

    def memoize(*args, **kwargs):
        def _wrap(f):
            return f
        return _wrap

from glibcparse.tokenize import Tokenizer
from glibcparse.preprocess import Screener
from glibcparse.parser import TranslitParser
from glibcparse.exc import LException

log = logbook.Logger('main')

def parse_translit(fn):
    return _parse_translit(os.path.normpath(os.path.abspath(fn)))

@memoize(100)
def _parse_translit(fn):
    with open(fn) as f:
        log.info("parse %s" % os.path.relpath(fn))
        tok = Tokenizer(Screener(f.read()))
        dirname = os.path.dirname(fn)

        def parse_func(new_fn):
            full_fn = os.path.join(dirname, new_fn)
            return parse_translit(full_fn)[0]

        p = TranslitParser(parse_func, tok)

        try:
            p.parse()
        except LException, e:
            log.critical("%s:%d.%d %s" % (
                fn,
                e.src.lineno,
                e.src.colno,
                e.err
            ))
            sys.exit(1)

        return p.ttbl, p


if '__main__' == __name__:
        import argparse
        parser = argparse.ArgumentParser(
            description="""Parses the provided files, calculates translation
                           tables and writes them as bzip'ed pickle files to
                           disc."""
        )
        parser.add_argument('files', metavar='FILE', nargs='+',
                            help="""localedata files to be parsed. Inside a
                            checkout of the glibc repository, these are usually
                            as [glibc]/localedata/locales/*""")
        parser.add_argument('--preprocess-only', '-E', action='store_true',
                            default=False, help="""Used to test comment
                            stripping and folding new lines.""")
        parser.add_argument('--debug', '-d', action='store_const',
                            dest='loglevel', const=logbook.DEBUG,
                            default=logbook.INFO, help="""Show debugging output
                            while working""")
        parser.add_argument('--output-dir', '-o',
                            default=os.path.join(
                                os.path.dirname(__file__),
                                'slugger/localedata'), help="""The directory
                            where generated output should be stored. Defaults
                            to slugger/localedata.""")
        parser.add_argument('--no-compression', '-C', dest='compress',
                            default=True, action='store_false', help="""Do not
                            compress resulting files.""")

        args = parser.parse_args()
        logbook.NullHandler().push_application()
        logbook.more.ColorizedStderrHandler(
            level=args.loglevel,
        ).push_application()

        if not has_memoize:
            log.error('You do not have remember (from PyPi) installed. This '\
                      'program will still work, but large jobs will run about'\
                      ' 10x slower')

        log.info('Storing output in %s' % args.output_dir)

        for fn in args.files:
            if args.preprocess_only:
                with open(fn) as f:
                    for c in Screener(f.read()):
                        sys.stdout.write(c)
            else:
                ttbl, parser = parse_translit(fn)
                if not parser.has_LC_IDENITIFCATION:
                    log.warning('No LC_IDENTIFICATION in "%s", skipping' %\
                                fn)
                else:
                    ext = '.ttbl' if not args.compress else '.ttbl.bz2'
                    out_fn = os.path.join(
                        args.output_dir,
                        os.path.basename(fn) + ext)
                    log.info('Writing output to %s' % out_fn)

                    outfile = open(out_fn, 'w') if not args.compress else\
                              bz2.BZ2File(out_fn, 'w', 1024**2, 9)
                    try:
                        pickle.dump(ttbl, outfile, pickle.HIGHEST_PROTOCOL)
                    finally:
                        outfile.close()



#!/usr/bin/env python
# coding=utf8

"""Main slugger module. See the docstring of Slugger for details."""

__version__ = '0.1'

import bz2
import imp
import re
import os
import sys

from remember.memoize import memoize
import pkg_resources
import unihandecode

from exc import LanguageNotFoundError
import languages
import languages.default_language

try:
    import cPickle as pickle
except ImportError:
    import pickle


_MEMOIZE = 20


_LANGCODE_RE = re.compile('^([a-z]+)(?:_([A-Z]+).*)?$')
@memoize(_MEMOIZE)
def _split_language(lang_code):
    m = _LANGCODE_RE.match(lang_code)

    if not m:
        raise LanguageNotFoundError('Bad language code: %r - expected format '
                                    'similiar to "en_US".' % lang_code)

    return m.groups()


@memoize(_MEMOIZE)
def _load_language(lang):
    language, territory = _split_language(lang)

    def _load_language_module(lang):
        mod_name = '%s.%s' % (languages.__name__, lang)

        if mod_name in sys.modules:
            return sys.modules[mod_name]

        mod = imp.load_module(
            mod_name,
            *imp.find_module(lang, languages.__path__)
        )

        return mod

    mod = None
    try:
        if territory:
            mod = _load_language_module('%s_%s' % (language, territory))
    except ImportError:
        pass

    if not mod:
        try:
            mod = _load_language_module('%s' % language)
        except ImportError:
            pass

    if not mod:
        raise LanguageNotFoundError('Could not find language module for %s' % \
                                    (lang,))

    return mod

@memoize(_MEMOIZE)
def _load_ttbl(lang):
    language, territory = _split_language(lang)

    language_resource = 'localedata/%s.ttbl' % language
    full_resource = 'localedata/%s_%s.ttbl' % (language, territory)
    fallback_resource = 'localedata/%s_%s.ttbl' % (language, language.upper())

    candidates = [
        full_resource + '.bz2',
        full_resource,
        language_resource + '.bz2',
        language_resource,
        fallback_resource + '.bz2',
        fallback_resource,
    ]

    for c in candidates:
        if pkg_resources.resource_exists(__name__, c):
            with pkg_resources.resource_stream(__name__, c) as f:
                buf = f.read()

            if c.endswith('.bz2'):
                buf = bz2.decompress(buf)

            return pickle.loads(buf)

    raise LanguageNotFoundError('Could not find translation table for %s' % \
                                lang)

@memoize(_MEMOIZE)
def _compile_rpl_exp(tbl):
    assert(tbl), "cannot compile empty table"

    exp = re.compile(u'(%s)' % u'|'.join(re.escape(k) for k in tbl.iterkeys()))

    subf = lambda m: tbl[m.group(1)]

    return lambda s: exp.sub(subf, s)


class Slugger(object):
    def __init__(self,
                 lang,
                 chain=None,
                 hanlang=None,
                 lowercase=True,
                 maxlength=100,
                 invalid_pattern='[^A-Za-z-]+',
                 invalid_replacement='-'):
        """Creates a new configuration for slugging.

        Note that creating Slugger instances can be expensive, if possible it
        should be avoided to repeat creating these everytime a slug is needed
        with the same configuration.

        :param lang: The language to use. Expects a language or
                     language_TERRITORY code, like 'de' or 'de_DE' (German,
                     German in Germany) or 'jp_JA' (Japanese in Japan).

                     If a language is not found, errors are silently swallowed
                     and the Slugger falls back to a basic default language.
        :param chain: A chain of operations. Each language comes with a default
                      chain, a list of strings. Upon instantiation,
                      init_CHAINNAME methods are called, while
                      :py:method:`sluggify` will call do_CHAINNAME to do its
                      work.
        :param hanlang: The language to use for transcribing chinese
                        characters, which are used in numerous asian languages.
                        This is mainly used when you have mixed-language input
                        and want better control, e.g. when having an English
                        text with Kanji mixed in and you want them interpreted
                        as Japanese, not Chinese.
        :param lowercase: Convert slug to lowercase..
        :param maxlength: Maximum length. Slugger will cut the slug short,
                          trying to keep words intact if possible.
        :param invalid_pattern: In the final step, this regular expression is
                                used to find invalid characters which are
                                replaced with *invalid-replacement*.
        :param invalid_replacement: The replacement for *invalid-pattern*-found
                                    characters.
        """
        self.lang = lang
        self.hanlang = hanlang
        self.lowercase = lowercase
        self.maxlength = maxlength
        self.invalid_pattern = invalid_pattern
        self.invalid_replacement = invalid_replacement

        # load the language
        try:
            self._lang = _load_language(lang)
        except LanguageNotFoundError:
            self._lang = languages.default_language

        self.substitution_tbl = getattr(
            self._lang, 'SUBSTITUTION',
            languages.default_language.SUBSTITUTION
        )

        self.chain = getattr(
            self._lang, 'CHAIN',
            languages.default_language.CHAIN)

        if chain:
            self.chain = chain

        # initialize chain
        for process in self.chain:
            getattr(self, 'init_%s' % process)()

    def do_subst(self, title):
        return self.subst_sub(title)

    def do_ttbl(self, title):
        return self.ttbl_sub(title)

    def do_unihandecode(self, title):
        return self.unihandecoder.decode(title)

    def do_cleanup(self, title):
        if self.lowercase:
            title = title.lower()

        words = title.split()
        if words:
            if len(words[0]) > self.maxlength:
                return words[0][:self.maxlength]

            total = 0
            collected = []
            while words:
                w = words.pop(0)
                if len(w) + total < self.maxlength:
                    total += 1 + len(w)
                    collected.append(w)

            title = ' '.join(collected)

        title = self.invalid_exp.sub(self.invalid_replacement, title)

        return title

    def init_subst(self):
        if self.substitution_tbl:
            self.subst_sub = _compile_rpl_exp(self.substitution_tbl)
        else:
            self.subst_sub = lambda s: s

    def init_ttbl(self):
        # load translation table
        self.ttbl = _load_ttbl(self.lang)
        self.ttbl_sub = _compile_rpl_exp(self.ttbl)

    def init_unihandecode(self):
        hanlang = self.hanlang or self.lang[0:2]

        # work-around for annoying bugs in unihandecode
        # author has been emailed
        if hanlang == 'ja':
            hanlang = "ja"
        elif hanlang == 'kr':
            hanlang = "kr"
        elif hanlang == 'vn':
            hanlang = "vn"

        self.unihandecoder = unihandecode.Unihandecoder(lang=hanlang)

    def init_cleanup(self):
        self.invalid_exp = re.compile(self.invalid_pattern)

    def sluggify(self, title):
        """Turn *title* into a slug.

        This will run the whole chain and return the end result of the
        transformation."""
        title = unicode(title)
        for process in self.chain:
            title = getattr(self, 'do_%s' % process)(title)

        return title
